# TODO Solve integration of process URI.
# Deprecated RemoteClient sends additional header with process URI.
# This is needed e.g. to be able to report progress or use locks.
# Process URI (or something similar) will be probably part of execution
# context, but until then commands which require process cannot be
# simply overwritten to use new command client (not without writing
# custom session handler).

require 'cgi'
require 'uu/os/http/http_client'
require 'uu/os/cmd/command_result'
require 'uu/os/cmd/command_handler'
require 'uu/os/cmd/session_handler'
require 'uu/os/util/name_converter'
require 'uu/os/lang/json_patch'
require 'uu/os/lang/exception_ext'
require 'uu/os/lang/future'
require 'uu/os/util/path_based_config'

module UU
  module OS
    module CMD

      # CommandClient is used for execution of command requests.
      # @example Basic usage:
      #    client = UU::OS::CMD::CommandClient.new('cds-bookcase', session)
      #    attributes = client.exec('Bookcase/getAttributes', bookcase_uri)
      class CommandClient
        include UU::OS::Util::NameConverter

        # Default base URL to be used for invocation of all commands.
        # @private
        DEFAULT_BASE_URL = 'https://cmd.plus4u.net' # :nodoc:
        private_constant :DEFAULT_BASE_URL

        # Default command handlers used for invocation of all commands
        # (to be used if no custom handlers are provided or if custom
        # handler does not know how to handle specific request/response).
        # @private
        DEFAULT_COMMAND_HANDLERS = [UU::OS::CMD::CommandHandler.new] # :nodoc:
        private_constant :DEFAULT_COMMAND_HANDLERS

        # Default session handler used for invocation of all commands
        # (to be used if no custom handler is provided).
        # @private
        DEFAULT_SESSION_HANDLER = UU::OS::CMD::SessionHandler.new # :nodoc:
        private_constant :DEFAULT_SESSION_HANDLER

        # Default action to be used with generic command invocation (invocation
        # using {UU::OS::CMD::CommandClient#invoke invoke} method).
        # @private
        DEFAULT_ACTION = :exec # :nodoc:
        private_constant :DEFAULT_ACTION

        # Default name of parameter in which to send URI of main object.
        # @private
        DEFAULT_MAIN_OBJECT_URI_PARAMETER_NAME = :uuUri # :nodoc:
        private_constant :DEFAULT_MAIN_OBJECT_URI_PARAMETER_NAME

        # Default name of parameter in which to send wrapped parameters
        # for GET requests.
        # @private
        DEFAULT_PARAMETERS_PARAMETER_NAME = :uuDto # :nodoc:
        private_constant :DEFAULT_PARAMETERS_PARAMETER_NAME

        # External command client configuration.
        # @private
        EXTERNAL_CONFIGURATION = UU::OS::Util::PathBasedConfig.new # :nodoc:
        private_constant :EXTERNAL_CONFIGURATION

        # Creates new instance of {CommandClient CommandClient} for given application (or application module).
        # @param [String] app_url_code Application URL code in format "+vendor+-+app+[-+module+]".
        # @param [Object] session Object providing execution environment data. All commands invoked from this instance will use
        #    data provided by this session object (even if other session, e.g. session containing other authentication, data was
        #    meanwhile created). For details of session object requirements see documentation of method
        #    {UU::OS::CMD::SessionHandler#handle_session handle_session}.
        # @param [Hash] options Additional configuration options.
        # @option options [String] :base_url Base URL common for invocation of all commands (defaults to https://cmd.plus4u.net).
        # @option options [TrueClass, FalseClass] :validate_ssl Flag if SSL communication should be validated (defaults to +true+
        #    which requires valid server certificate).
        # @option options [String] :ssl_version SSL version to be used for communication (defaults to value provided by particular
        #    version of Ruby).
        # @option options [String] :ssl_cert_store Path to alternate SSL certificate store to be used instead of default one.
        # @option options [TrueClass, FalseClass] :reduce_network_usage Flag if network usage should be optimized in tradeoff for
        #    reasonable CPU/memory performance hit (defaults to +true+).
        # @option options [Symbol, String] :default_action Default use case action to be used with generic command invocation
        #    (invocation using {UU::OS::CMD::CommandClient#invoke invoke} method, defaults to +exec+).
        # @option options [#handle_request, #handle_response, #handle_error, #handle_redirect,
        #    Array<#handle_request, #handle_response, #handle_error, #handle_redirect>] :command_handlers Any object (or list of
        #    objects) that responds to {CMD::CommandHandler#handle_request handle_request}, {CMD::CommandHandler#handle_response
        #    handle_response}, {CMD::CommandHandler#handle_error handle_error} or {CMD::CommandHandler#handle_redirect handle_redirect}.
        #    In case this parameter is defined, given handlers are used additionally to default handler (which is always invoked last).
        #    In case of list, order of handler invocation is same as order in list.
        # @option options [#handle_session] :session_handler Any object that responds to method {CMD::SessionHandler#handle_session
        #    handle_session}. In case this parameter is defined, given object is used as session handler instead of default one.
        # @option options [Fixnum] :connect_timeout Timeout for remote connection initialization in seconds (defaults to 1 second,
        #    use 0 for infinite timeout).
        # @option options [Fixnum] :request_timeout Request timeout in seconds (defaults to 30 seconds, use 0 for infinite timeout).
        # @option options [TrueClass, FalseClass] :append_routing_data If set to +true+ (default value) routing data (in particular
        #    territory identification) are added to request path.
        def initialize(app_url_code, session, options = nil)
          use_legacy_config = get_param(options, :use_legacy_config) || false
          @app_url_code = app_url_code
          @session = session
          @command_handlers = DEFAULT_COMMAND_HANDLERS.dup
          command_handlers = get_param(options, :command_handlers, :classes)
          @command_handlers.unshift(*command_handlers) if command_handlers
          session_handler = get_param(options, :session_handler, :classes)
          if session_handler.respond_to?:handle_session
            @session_handler = session_handler
          else
            @session_handler = DEFAULT_SESSION_HANDLER
          end
          @default_action = get_param(options, :default_action, :string)
          @default_action = DEFAULT_ACTION if @default_action.nil?
          @main_object_uri_parameter_name = get_param(options, :main_object_uri_parameter_name, :string)
          @main_object_uri_parameter_name = DEFAULT_MAIN_OBJECT_URI_PARAMETER_NAME if @main_object_uri_parameter_name.nil?
          @parameters_parameter_name = get_param(options, :parameters_parameter_name, :string)
          @parameters_parameter_name = DEFAULT_PARAMETERS_PARAMETER_NAME if @parameters_parameter_name.nil?
          @append_routing_data = get_param(options, :append_routing_data, :boolean)
          @append_routing_data = true if @append_routing_data.nil?
          # Prepare instance of HTTPClient which is used for communication
          if use_legacy_config
            base_url = get_param(options, :base_url, :string, :'server-url')
            validate_ssl = get_param(options, :validate_ssl, :boolean, :'validate-server-cert')
            ssl_version = get_param(options, :ssl_version, :string, :'server-ssl-version')
          else
            base_url = get_param(options, :base_url, :string)
            validate_ssl = get_param(options, :validate_ssl, :boolean)
            ssl_version = get_param(options, :ssl_version, :string)
          end
          base_url ||= get_param(options, :default_base_url) || DEFAULT_BASE_URL
          @http_client = UU::OS::HTTP::HTTPClient.new(base_url)
          # Reconfigure default headers
          @http_client.set_header(:accept, 'application/json', '*/*')
          @http_client.set_header(:content_type, 'application/json')
          # Reconfigure client based on options parameter
          @http_client.tranparent_gzip_decompression = get_param(options, :reduce_network_usage, :boolean)
          @http_client.ssl_validation = validate_ssl
          @http_client.ssl_version = ssl_version
          @http_client.ssl_cert_store = get_param(options, :ssl_cert_store, :string)
          @http_client.connect_timeout = get_param(options, :connect_timeout, :number)
          @http_client.request_timeout = get_param(options, :request_timeout, :number)
        end

        # Allows to globally register additional command handler (e.g. module
        # can register handler to process its exceptions). Handler is always
        # added on top of actual handlers list (so system default handler remains
        # always last in chain).
        # @note This registration should be used only by uuOS modules!
        # @private
        def self.register_default_command_handler(handler) # :nodoc:
          if !DEFAULT_COMMAND_HANDLERS.include?(handler)
            DEFAULT_COMMAND_HANDLERS.unshift(handler)
          end
        end

        # Allows to globally replace default session handler (e.g. for handler
        # which is aware of classes and data types not known in scope of this
        # library).
        # @note This registration should be used only by uuOS modules!
        # @private
        def self.default_session_handler=(handler) # :nodoc:
          self.send(:remove_const, :DEFAULT_SESSION_HANDLER) if self.const_defined?:DEFAULT_SESSION_HANDLER
          self.const_set(:DEFAULT_SESSION_HANDLER, handler)
        end

        # Allows to globally reconfigure base URL which might be usable for
        # testing environment.
        # @private
        def self.default_base_url=(base_url) # :nodoc:
          self.send(:remove_const, :DEFAULT_BASE_URL) if self.const_defined?:DEFAULT_BASE_URL
          self.const_set(:DEFAULT_BASE_URL, base_url)
        end

        # Loads external configuration to be used by all instances of command client. This
        # configuration has priority over implicit default values, however can be overridden
        # by command client {UU::OS::CMD::CommandClient#initialize initialization options}.
        # Configuration is "path based", therefore it is possible to globally reconfigure
        # behavior only for certain modules or applications, leaving others with default
        # values. Repeated invocation of method does not overwrite previously loaded values,
        # but configurations are merged. Therefore it is possible for each module to load
        # its configuration (as long as they are using unique path).
        #
        # Configuration keys are expected in format +<dot.separated.path>.<option>+ where
        # +path+ is based on application module URL path (e.g. +uu.os.artifact+ or +cds.issues+)
        # and +option+ is one of {UU::OS::CMD::CommandClient#initialize initialization options}.
        # @param [String, #read, Hash] configuration Possible values are:
        #
        # * +String+ - Path to configuration file in properties format.
        # * +#read+ - Instance of existing +IO+ (e.g. +File+) pointing to data in properties format.
        # * +Hash+ - Configuration in plain key-value format.
        #
        # @return [void]
        def self.load_configuration(configuration)
          EXTERNAL_CONFIGURATION.merge!(configuration)
        end

        # Invokes command with given parameters. When using this method, default use case action is invoked (which may
        # be configured via +options+ parameter in {#initialize CommandClient constructor} and defaults to +exec+).
        # @param [String] cmd_url_path URL path of command to be invoked
        # @param [URI, String] uu_uri URI of main object (object to be processed by invoked command) given either
        #    as plain string or instance of URI (any other type is transformed using {Object#to_s #to_s} method).
        # @param [Hash] parameters Additional execution parameters.
        # @option parameters [Hash, IO, String] :parameters Actual method parameters (typically represented as DTO). Note
        #    that command data interchange protocol uses "camelCase" for attribute names. In case "snake_case" attributes
        #    are given in Hash parameter, they might be ignored by server (no transformation of attribute names is done by
        #    CommandClient). Particular allowed type of parameter depends on additional options and command type:
        #
        #    * +POST commands+
        #      * +no options+ - Hash is transformed to JSON. String or IO are send without transformation (but expected
        #        content is valid JSON representing serialized Hash).
        #      * +:binary_upload+ - Hash containing at least +:data+ attribute (Attribute may contain String or IO and is always
        #        send as last part). Additionally Hash may contain +name+, +contentType+ and +size+ parameters. If not set, these
        #        parameters are derived from +data+ parameter where possible.
        #      * +:multipart_upload+ - Hash in plain key-value structure (any nested structure is transformed to JSON)
        #
        #    * +GET commands+
        #      * Only allowed type is Hash which is transformed to query parameters (nested structures are transformed to
        #        JSON, however it is strongly recommended to not use them because of limited length of request URL).
        #
        # @option parameters [TrueClass, FalseClass, nil] :stream_response Flag if response should be streamed - returned as IO
        #    (defaults to +false+, should be set to +true+ in case response might be bigger than 5MB).
        # @option parameters [String] :content_type Content type of request (defaults to application/json or multipart/form-data
        #    see description of +:parameters+ for details).
        # @option parameters [String, nil] :accept Requested content type of response (defaults to application/json).
        # @option parameters [Symbol, String, nil] :invocation_method Type of command invocation (defaults to POST).
        # @option parameters [TrueClass, FalseClass, nil] :async_invocation Flag if command should be invoked asynchronously
        #    on client. Note that this does not mean invocation of asynchronous command! Result is returned as instance
        #    of {UU::OS::Lang::Future Future}. (defaults to +false+)
        # @option parameters [TrueClass, FalseClass, nil] :raw_response Flag if {CMD::CommandResult CommandResult} object should
        #    be returned instead of actual processed result (defaults to +false+, if set to +true+ {CMD::CommandResult
        #    CommandResult} is returned also from asynchronous invocation).
        # @option parameters [#handle_request, #handle_response, #handle_error, #handle_redirect,
        #    Array<#handle_request, #handle_response, #handle_error, #handle_redirect>, nil] :command_handlers Any object (or list of
        #    objects) that responds to {CMD::CommandHandler#handle_request handle_request}, {CMD::CommandHandler#handle_response
        #    handle_response}, {CMD::CommandHandler#handle_error handle_error} or {CMD::CommandHandler#handle_redirect handle_redirect}.
        #    In case this parameter is defined, given handlers are used additionally and prior to handlers declared in {#initialize
        #    CommandClient constructor}. In case of list, order of handler invocation is same as order in list.
        # @option parameters [TrueClass, FalseClass, nil] :binary_upload Flag if binary value is being send as command parameter (defaults to +false+).
        # @option parameters [TrueClass, FalseClass, nil] :multipart_upload Flag if parameters should be transformed into multipart message (defaults
        #    to +false+).
        # @option parameters [Object, nil] :callback Object containing parameters for callback (defines operation to be invoked once asynchronous
        #    command is finished). Callbacks are supported only for server side where valid callback parameters to be used can be obtained via method
        #    {UU::OS::CMD::ExecutionContext#prepare_callback}.
        # @option parameters [Object, nil] :... Any given options are not modified. Therefore custom command handler can define additional options
        #    to be used for processing command request or command response.
        # @return [Object, UU::OS::CMD::CommandResult, UU::OS::Lang::Future] Command result. Actual type of returned value depends on given
        #    options and also on set of used command handlers. Note that in case of Hash result value, all attribute names will be in "camelCase"
        #    since this notation is by default used in command data interchange protocol (and no transformation of attribute names is done
        #    by CommandClient).
        def invoke(cmd_url_path, uu_uri, parameters = nil)
          return __invoke(cmd_url_path, @default_action, uu_uri, parameters)
        end

        # Returns wrapped instance of configured HTTPClient (to be able to invoke
        # non-standard command requests without need to fully configure own instance
        # of HTTPClient).
        # @return [UU::OS::HTTP::HTTPClient] Instance of HTTPClient
        # @private
        def unwrap() # :nodoc:
          @http_client
        end

        # Wraps raw HTTPResponse into CommandResult, so that it is possible to use
        # command handlers even for non-standard command requests.
        # @param [UU::OS::HTTP::HTTPResponse] http_response Raw HTTP response.
        # @param [Hash] options Additional parameters to be used by handlers (see
        #    {#invoke #invoke for possible values}. Mandatory parameters are
        #    :invocation_method and :main_object_uri_parameter_name which cannot be
        #    obtained from response and are unknown for generic command request which
        #    we are wrapping.
        # @return [UU::OS::CMD::CommandResult] Command result.
        # @private
        def wrap_result(http_response, options) # :nodoc:
          if options.nil? || !options.has_key?(:invocation_method) || !options.has_key?(:main_object_uri_parameter_name)
            raise ArgumentError, "Missing options with parameters :invocation_method and :main_object_uri_parameter_name."
          end
          # Prepare handlers
          handler_chain = @command_handlers.dup
          custom_handlers = get_param(options, :command_handlers)
          handler_chain.unshift(*custom_handlers) if custom_handlers
          # Prepare options
          options = (options.nil?) ? {} : options.dup
          options.delete(:parameters)
          options[:command_handlers] = handler_chain
          # Wrap HTTP response
          UU::OS::CMD::CommandResult.new(http_response, handler_chain, options)
        end

        # Returns base URL for which is this instance of client configured.
        # @return [String] Command client base URL.
        # @private
        def base_url # :nodoc:
          return @http_client.instance_variable_get(:@base_url)
        end

        # Override inspect to hide internal structures.
        # @private
        def inspect # :nodoc:
          "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)}>"
        end

        private

        # Returns parameter value from given set of parameters, external
        # configuration or nil if parameter is not available.
        # @private
        def get_param(params, name, load_from_ext = nil, alt_ext_name = nil) # :nodoc:
          raise TypeError, "no implicit conversion of #{params.class} to Hash" if !params.nil? && !params.respond_to?(:[])

          value = params[name] if !params.nil? && params.has_key?(name)
          return value if !value.nil? || load_from_ext.nil?

          # If defined, load parameter from external configuration under alternative name
          # (used to be able to process "legacy" configuration parameters).
          value = EXTERNAL_CONFIGURATION.get(alt_ext_name.nil? ? name : alt_ext_name, @app_url_code)
          return value if value.nil?
          begin
            case load_from_ext
            when :number
              if value =~ /^[+-]?[0-9]+$/
                value = value.to_i
              elsif value =~ /^[+-]?[0-9]+[.]?[0-9]+(e[+-][0-9]+)?$/i
                value = value.to_f
              else
                raise
              end
            when :boolean
              raise if value !~ /^(true|false)$/i
              value = !!(value =~ /^true$/i)
            when :classes
              classes = []
              value.split(',').each do |cls|
                cls = cls.strip
                classes << Object.const_get(cls) if !cls.empty?
              end
              value = (classes.size <= 1 ? classes[0] : classes)
            end
          rescue Exception
            raise TypeError, "unable to convert \"#{value}\" of parameter \"#{name}\" to #{load_from_ext}"
          end
          return value
        end

        # Transforms call of unknown method to specific action.
        # @private
        def method_missing(method, *args, &block) #:nodoc:
          # Command invocation must have method, entity and optional parameters
          if (args.size < 2) || (args.size > 3)
            return super
          else
            return __invoke(args[0], to_camel_case(method), args[1], args[2])
          end
        end

        # Internal method for actual command invocation. It is prefixed by double underscore
        # so that it prevents confusion in case "invoke" is used as use case action (which
        # may happen since "invoke" is nice name for action).
        # @see #invoke for parameter description.
        # @private
        def __invoke(cmd_url_path, action, uu_uri, parameters) # :nodoc:
          # Prepare handlers
          handler_chain = @command_handlers.dup
          custom_handlers = get_param(parameters, :command_handlers)
          handler_chain.unshift(*custom_handlers) if custom_handlers
          # Prepare options (to be used mainly for result processing)
          options = (parameters.nil?) ? {} : parameters.dup
          options.delete(:parameters)
          # Use explicitly defined HTTP method or default value
          options[:invocation_method] ||= :post
          if @append_routing_data == true
            # TODO Rewrite to using uuURI once library is implemented
            # (we cannot use UESURI from uu_os_framework due to dependency loop)
            if uu_uri.to_s =~ /[^:]+:([^\[:]*)(?:\[([^\]]+)\])?:.*/
              # Territory code must be URL-escaped (because it may contain invalid character, namely slash)
              http_method = @http_client.send("prepare_#{options[:invocation_method].to_s.downcase}", $2 || CGI.escape("#{$1}"), @app_url_code, cmd_url_path, action)
            else
              http_method = @http_client.send("prepare_#{options[:invocation_method].to_s.downcase}", @app_url_code, cmd_url_path, action)
            end
          else
            http_method = @http_client.send("prepare_#{options[:invocation_method].to_s.downcase}", @app_url_code, cmd_url_path, action)
          end
          # Override default content-type and accept headers if defined
          content_type = get_param(options, :content_type)
          accept = get_param(options, :accept)
          http_method.set_header(:content_type, content_type) if content_type
          http_method.set_header(:accept, accept) if accept
          # Set main main_object_uri parameter
          options[:main_object_uri_parameter_name] ||= @main_object_uri_parameter_name
          http_method.set_parameter(options[:main_object_uri_parameter_name], uu_uri.to_s)
          # Process session
          @session_handler.handle_session(@session, http_method.headers, http_method.parameters)
          # Add all used handlers to options (so that handlers know about others)
          options[:command_handlers] = handler_chain
          # Add session (so that handlers know about user who invoked command)
          options[:session] = @session
          # Process command parameters (data)
          params = get_param(parameters, :parameters)
          # Transform callback (move it from options to uuCtx parameter)
          callback = extract_callback(parameters)
          if !callback.nil?
            params = params.nil? ? {} : params.dup
            params[:uuCtx] ||= {}
            params[:uuCtx][:callback] = callback
          end
          # Process payload by handlers
          payload = process_payload(handler_chain, params, http_method.headers, http_method.parameters, options)
          # Optionally transform payload for get requests (wrap to single data parameter)
          if options[:invocation_method] =~ /^get$/i && !@parameters_parameter_name.nil? && !@parameters_parameter_name.empty?
            payload = {@parameters_parameter_name => payload}
          end

          # Asynchronous request must be issued also in order to stream response
          if get_param(parameters, :stream_response) || get_param(parameters, :async_invocation)
            raw_result = http_method.execute_async(payload)
          else
            raw_result = http_method.execute(payload)
          end
          result = UU::OS::CMD::CommandResult.new(raw_result, handler_chain, options)

          if get_param(parameters, :raw_response)
            return result
          elsif get_param(parameters, :async_invocation)
            return UU::OS::Lang::Future.new(result)
          else
            return result.get
          end
        end

        # Delegates processing of request payload to set of command handlers.
        # @private
        def process_payload(handler_chain, payload, http_headers, http_parameters, options) # :nodoc:
          handler_chain.each do |handler|
            next if !handler.respond_to?:handle_request
            res = handler.handle_request(payload, http_headers, http_parameters, options)
            raise res if res.is_a?Exception
            return res if res != CMD::CommandHandler::NEXT
          end
          raise TypeError, "Unable to process payload of type #{payload.class} using handlers #{handler_chain}."
        end

        # Extracts callback from given parameters and transforms it to Hash
        # which can be sent on uuCtx
        # @private
        def extract_callback(parameters) # :nodoc:
          return if !parameters.kind_of?(Hash) || parameters[:callback].nil?
          callback = parameters[:callback].dup
          callback[:appUrlCode] ||= callback.delete(:app_url_code)
          callback[:cmdUrlPath] ||= callback.delete(:cmd_url_path)
          callback[:mainObjectUri] ||= callback.delete(:uu_uri)
          session = callback.delete(:session)
          if session.kind_of?Hash
            callback[:authToken] = session[:accessToken]
          elsif session.respond_to?(:access_token)
            callback[:authToken] = session.access_token
          elsif session.kind_of?(String)
            callback[:authToken] = session
          end
          return callback
        end

      end

    end
  end
end
