require 'uri'
require 'uu/os/cmd/command_executor'
require 'uu/os/cmd/command_result'
require 'uu/os/http/http_headers'
require 'uu/os/cmd/async_command_state'

module UU
  module OS
    module CMD

      # Future actor providing result from invocation of asynchronous command.
      # @private
      class CommandFutureActor # :nodoc:

        # Creates new instance of command future actor.
        # @param [Hash] future Hash representing command future.
        # @param [Hash] options Options to be used for retrieving
        #    asynchronous command status (by default same options
        #    used for command invocation should be used in order
        #    to apply same set of handlers, parameter names and
        #    session).
        def initialize(future, options = nil)
          if !self.class.command_future?(future)
            raise ArgumentError.new('Given object is not valid structure')
          end
          @future = future
          @options = options
          @client = UU::OS::CMD::CommandClient.new(future[:appUrlCode], options[:session],
            default_action: "#{future[:action]}AsyncStatus")
        end

        # Returns result of asynchronous command.
        # @return [Object] Command result.
        def get
          wait = 0.5
          while true
            # Obtain asynchronous task result...
            result = @client.invoke(@future[:cmdUrlPath], @future[:uuUri], parameters: {taskUri: @future[:taskUri]})
            if result[:state] != UU::OS::CMD::AsyncCommandState::WAITING && result[:state] != UU::OS::CMD::AsyncCommandState::RUNNING
              # Create mock response
              if result[:state] == UU::OS::CMD::AsyncCommandState::ERROR
                exception = result[:exception]
                if exception =~ /"status":([^,}]+)/
                  status = $1.to_i
                else
                  status = 500
                end
                response = AsyncCmdResponse.new(status, exception)
              else
                response = AsyncCmdResponse.new(200, result[:result].to_json)
              end
              # Wrap future result to command result to be able to apply all handlers
              # (especially necessary for error handling to be able to throw custom
              # exceptions declared by command).
              result = UU::OS::CMD::CommandResult.new(response, @options[:command_handlers], @options)
              return result.get
            end
            wait *= 2
            sleep(wait)
          end
          return result
        end

        # Provides serialization to JSON.
        # @return [String] Serialized command future.
        def to_json(*args)
          return @future.to_json(*args)
        end

        # Override inspect to hide internal structures.
        # @private
        def inspect # :nodoc:
          "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)}>"
        end

        # Checks if given object represents valid command future.
        # @param [Object] future Object to be checked
        # @return [TrueClass, FalseClass] Check result
        # @private
        def self.command_future?(future) # :nodoc:
          return future.kind_of?(Hash) && future.size == 6 && future[:future] == true && future.has_key?(:taskUri) &&
            future.has_key?(:appUrlCode) && future.has_key?(:cmdUrlPath) && future.has_key?(:action) && future.has_key?(:uuUri)
        end

        private

        # Mock {UU::OS::HTTP::HTTPResponse} object used to wrap
        # asynchronous invocation result so that it can be processed by
        # command handlers.
        # @private
        class AsyncCmdResponse # :nodoc:

          attr_reader :status

          attr_reader :headers

          attr_reader :body

          attr_reader :request

          def initialize(status, body)
            @headers = UU::OS::HTTP::HTTPHeaders.new(content_type: 'application/json')
            @status = status
            @body = body
          end

          def inspect # :nodoc:
            "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)}>"
          end

        end

      end

    end
  end
end
