require 'uu/os/util/url_converter'
require 'uu/os/util/configuration_loader'

module UU
  module OS
    module Util

      # Object providing configuration based not only on configuration
      # parameter name, but also on path.
      # @private
      class PathBasedConfig # :nodoc:
        include UU::OS::Util::URLConverter
        include UU::OS::Util::ConfigurationLoader

        # Creates new instance of path based configuration object.
        # @param [String, #read, Hash] source Path to configuration file instance of
        #   IO providing configuration data or Hash with existing configuration.
        def initialize(source = nil)
          @config = load(source)
        end

        # Merges another configuration into this configuration object.
        # @param [String, #read, Hash] source Path to configuration file instance of
        #   IO providing configuration data or Hash with existing configuration.
        def merge!(source)
          @config.merge!(load(source))
        end

        # Returns value for +key+ and given +path+.
        # @param [String, Symbol] name Configuration parameter name
        # @param [Array<String>] path Service path.
        # @return [String, nil] Configuration value.
        def get(name, *path)
          return nil if @config.nil? || @config.empty?
          name = name.to_s.downcase
          path = concat_path(*path).downcase
          # Transform legacy paths to match CMD_v3 paths:
          # 1) Remove dedication prefixes ("ues/wcp" or "cXY" or "secXY")
          path.gsub!(/^(\/)?((ues\/wcp)|(c[0-9]+)|(sec[0-9]+))\//, '')
          # 2) Transform "ues/???/" path to "uu/os" path
          path.gsub!(/^(\/)?ues\/[a-z]+\//, 'uu/os/')
          # Transform path to configuration prefix
          # (replace separators by dots)
          path.gsub!(/(^\/)|(\/$)/, '')
          path.gsub!(/[\/\-_]/, '.')
          # Look for configuration value from full path to default value.
          key = "#{path}.#{name}"
          while !path.empty? && !@config.has_key?(key)
            path = path.split('.')[0..-2].join('.')
            key = "#{path}.#{name}"
          end
          return path.empty? ? @config[name] : @config[key]
        end

        private

        # Loads configuration from source.
        # @param [String, #read, Hash] source Configuration source
        # @return [Hash] Loaded configuration.
        # @private
        def load(source) # :nodoc:
          config = {}
          if source.kind_of?(Hash)
            config = source.dup
          elsif source.kind_of?self.class
            config = source.instance_variable_get(:@config).dup
          elsif !source.nil?
            config = load_configuration(source)
          end
          # In case original uu-client.properties are loaded, transform
          # known parameters to maintain at least some compatibility.
          compat_config = {}
          config.each do |key, value|
            case key
            # New command client uses only single SSL validation parameter. When old parameters are found
            # change them to new one (use dash instead of underscore as separator to distinguish between
            # legacy configuration and new command client configuration).
            when /\.trust-server-cert$|^trust-server-cert$/
              compat_config[key.gsub(/trust-server-cert$/, 'validate-server-cert')] = "#{!!!(value.to_s =~ /^true$/i)}"
            when /\.trust-server-hostname$|^trust-server-hostname$/
              compat_config[key.gsub(/trust-server-hostname$/, 'validate-server-cert')] = "#{!!!(value.to_s =~ /^true$/i)}"
            # To maintain same style of configuration as for legacy commands, allow to use same configuration
            # keys with "cmd-" prefix. These parameters are converted to "options" format used by new command
            # client. 
            when /\.cmd-server-url$|^cmd-server-url$/
              compat_config[key.gsub(/cmd-server-url$/, 'base_url')] = value
            when /\.cmd-append-routing-data$|^cmd-append-routing-data$/
              compat_config[key.gsub(/cmd-append-routing-data$/, 'append_routing_data')] = value
            when /\.cmd-server-ssl-version$|^cmd-server-ssl-version$/
              compat_config[key.gsub(/cmd-server-ssl-version$/, 'ssl_version')] = value
            when /\.cmd-trust-server-cert$|^cmd-trust-server-cert$/
              compat_config[key.gsub(/cmd-trust-server-cert$/, 'validate_ssl')] = "#{!!!(value.to_s =~ /^true$/i)}"
            when /\.cmd-trust-server-hostname$|^cmd-trust-server-cert$/
              compat_config[key.gsub(/cmd-trust-server-hostname$/, 'validate_ssl')] = "#{!!!(value.to_s =~ /^true$/i)}"
            end
          end
          config.merge!(compat_config)
          return config
        end

      end

    end
  end
end
